
// The following ifdef block is the standard way of creating macros which make exporting 
// from a DLL simpler. All files within this DLL are compiled with the GXDLL_EXPORTS
// symbol defined on the command line. this symbol should not be defined on any project
// that uses this DLL. This way any other project whose source files include this file see 
// GXDLL_API functions as being imported from a DLL, wheras this DLL sees symbols
// defined with this macro as being exported.

#ifndef _GX_H
#define _GX_H

#include <w32std.h>

#ifdef GXDLL_EXPORTS
#define GXDLL_API __declspec(dllexport)
#else
#define GXDLL_API __declspec(dllimport)
#endif

typedef unsigned long DWORD;
typedef unsigned long HWND;

struct _POINT {
	int x;
    int y;
};

struct GXDisplayProperties {
	DWORD cxWidth;
	DWORD cyHeight;			// notice lack of 'th' in the word height.
	long cbxPitch;			// number of bytes to move right one x pixel - can be negative.
	long cbyPitch;			// number of bytes to move down one y pixel - can be negative.
	long cBPP;				// # of bits in each pixel
	DWORD ffFormat;			// format flags.
};

struct GXKeyList {
	short vkUp;				// key for up
	struct _POINT ptUp;				// x,y position of key/button.  Not on screen but in screen coordinates.
	short vkDown;
	struct _POINT ptDown;
	short vkLeft;
	struct _POINT ptLeft;
	short vkRight;
	struct _POINT ptRight;
	short vkA;
	struct _POINT ptA;
	short vkB;
	struct _POINT ptB;
	short vkC;
	struct _POINT ptC;
	short vkStart;
	struct _POINT ptStart;
};

struct GXScreenRect {
	DWORD dwTop;
	DWORD dwLeft;
	DWORD dwWidth;
	DWORD dwHeight;
};

enum GXDirScrAccEngPanics
    {
    GXDirScrAccEngAlreadyStarted = 1,
	GXDirScrAccEngAlreadyStopped
    };

//

// CLASS DECLARATION


class CGXDirScrAcc : public MDirectScreenAccess	
	{

public:

	// Construction & destruction
	static CGXDirScrAcc* NewL 
		(
		RWsSession& aClient, 
		CWsScreenDevice& aScreenDevice, 
		RWindow& aWindow,
        TBool aUseFrameBuffer=EFalse	
		);

	~CGXDirScrAcc();	

public:

	// Start and stop the drawing
	void StartDrawingL();
	void StopDrawing();

	// Are we drawing
	inline TBool Drawing(){return iDrawing;}
	void UpdateL();
	void DoCancel();
    void Resume();
    void Suspend();
    void PumpEvents();
private:

	// Implement MDirectScreenAccess 
	void Restart(RDirectScreenAccess::TTerminationReasons aReason);
	void AbortNow(RDirectScreenAccess::TTerminationReasons aReason);
	TInt HandleWsEvent(const TWsEvent& aEvent);
public: 
      TPoint iMouseCursor;
      TBool  isCursorVisible;
private:

    // private constructors - use NewL to construct a CDirScrAccEng!
    CGXDirScrAcc
        (
		RWsSession& aClient, 
		CWsScreenDevice& aScreenDevice, 
		RWindow& aWindow,
        TBool aUseFrameBuffer        
        );

    void ConstructL();	

private:

	// Window server handling
	RWsSession& iClient;
public:	
	CWsScreenDevice& iScreenDevice;
	RWindow& iWindow;
	TBool isSuspend;
private:
	// Direct Screen Access
	CDirectScreenAccess* iDirectScreenAccess;
	CFbsBitGc* iGc;
	RRegion* iRegion;

	// Positional plus directional information
	TPoint iPosition;	

	// Are we drawing
	TBool iDrawing;
	TSize ScreenSize;	

    TUint8* iScreenAddr;        // frame buffer address
  public:  
    CFbsBitmap* iOffScreenBmp;  // offscreen bitmap for double-buffered drawing
    TRect iFlagRect;
    TInt  iFrameBufLen;         // length of the frame buffer in bytes

    TBool iUseFrameBuffer;
    
    TRawEvent iRedraw;	        // raw event for forcing the update of screen 
	};

// Result is FullScreen display with 64K colors
int GXOpenDisplay(RWsSession& aClient, 
                    		CWsScreenDevice& aScreenDevice, 
                    		RWindow& aWindow);
int GXCloseDisplay();
void * GXBeginDraw();
int GXEndDraw();
int GXOpenInput();
int GXCloseInput();
//The following two lines modified by Dan East to make this header C compatible:
//Added "struct" to the following two prototypes:
struct GXDisplayProperties GXGetDisplayProperties();
struct GXKeyList GXGetDefaultKeys(int iOptions);
int GXSuspend();
int GXResume();
int GXSetViewport( DWORD dwTop, DWORD dwHeight, DWORD dwReserved1, DWORD dwReserved2 );
int GXIsDisplayDRAMBuffer();
int GXDrawing();

// Although these flags can be unrelated they still
// have unique values.

#define GX_FULLSCREEN	0x01		// for OpenDisplay()
#define GX_NORMALKEYS	0x02
#define GX_LANDSCAPEKEYS	0x03

	#define kfLandscape	0x8			// Screen is rotated 270 degrees
	#define kfPalette	0x10		// Pixel values are indexes into a palette
	#define kfDirect	0x20		// Pixel values contain actual level information
	#define kfDirect555	0x40		// 5 bits each for red, green and blue values in a pixel.
	#define kfDirect565	0x80		// 5 red bits, 6 green bits and 5 blue bits per pixel
	#define kfDirect888	0x100		// 8 bits each for red, green and blue values in a pixel.
	#define kfDirect444	0x200		// 4 red, 4 green, 4 blue
	#define kfDirectInverted 0x400

#endif