//Fixed Point Math Routines
//Copyright 2/2/01 Dan East
//Copyright 23.07.06 Dmitry Gazko
// partially used: libmad - MPEG audio decoder library
// Copyright (C) 2000-2004 Underbit Technologies, Inc.

#include "quakedef.h"
#include <math.h>

#if defined(DEBUG)&&defined(_X86_)
#include "windows.h"
#define FPM_VALIDATE
#endif

long long cnt_fromfloat;
long long cnt_tofloat;
long long cnt_todouble;
long long cnt_fromlong;
long long cnt_tolong; 
long long cnt_add;
long long cnt_sub;
long long cnt_mul;
long long cnt_div;
long long cnt_sin;
long long cnt_cos;
long long cnt_floor;
long long cnt_ceil;
long long cnt_abs;
long long cnt_tan;
long long cnt_atan;
long long cnt_sqrt;


void fpm_PrintfCounters()
{ Sys_Printf("FromFloat: %d\n",cnt_fromfloat);
  Sys_Printf("ToFloat: %d\n",cnt_tofloat);	
  Sys_Printf("ToDouble: %d\n",cnt_todouble);
  Sys_Printf("FromLong: %d\n",cnt_fromlong);
  Sys_Printf("ToLong: %d\n",cnt_tolong);
  Sys_Printf("Add: %d\n",cnt_add);
  Sys_Printf("Sub: %d\n",cnt_sub);
  Sys_Printf("Mul: %d\n",cnt_mul);
  Sys_Printf("Div: %d\n",cnt_div);
  Sys_Printf("Sin: %d\n",cnt_sin);
  Sys_Printf("Cos: %d\n",cnt_cos);
  Sys_Printf("Tan: %d\n",cnt_tan);
  Sys_Printf("ATan: %d\n",cnt_atan);
  Sys_Printf("Floor: %d\n",cnt_floor);
  Sys_Printf("Ceil: %d\n",cnt_ceil);
  Sys_Printf("Abs: %d\n",cnt_abs);
  Sys_Printf("Sqrt: %d\n",cnt_sqrt);
 }

void fpm_InitCounters()
{ cnt_fromfloat=0;
  cnt_tofloat=0;
  cnt_todouble=0;	
  cnt_fromlong=0;
  cnt_tolong=0;
  cnt_add=0;
  cnt_sub=0;
  cnt_mul=0;
  cnt_div=0;
  cnt_sin=0;
  cnt_cos=0;
  cnt_floor=0;
  cnt_ceil=0;
  cnt_abs=0;
  cnt_tan=0;
  cnt_atan=0;
  cnt_sqrt=0;	 
  Sys_Printf("Frame begin\n");
 }

 fixedpoint_t fpm_FromFloat(double f) {
// 	cnt_fromfloat++;
	return (fixedpoint_t)(f*4294967296.0);
}


double fpm_ToFloat(fixedpoint_t fxp) {
//	cnt_tofloat++;
	return ((((double)(fxp))/4294967296.0));
}

double fpm_ToDouble(fixedpoint_t fxp) {
//	cnt_todouble++;
	return ((((double)(fxp))/4294967296.0));
}


fixedpoint_t fpm_FromLong(long l) {
//   cnt_fromlong++;
   if (l<0) 
      return -(((fixedpoint_t)(-l))<<32);
   else
     return ((fixedpoint_t)(l))<<32;
}

long fpm_ToLong(fixedpoint_t fxp) {
//	cnt_tolong++; 
	if (fxp<0) 
	  return -((long)((-fxp)>>32));  
	else
	  return (long)(fxp>>32);   
}


 fixedpoint_t fpm_Add(fixedpoint_t fxp1, fixedpoint_t fxp2) {
// 	cnt_add++;
	return fxp1+fxp2;
}

 fixedpoint_t fpm_Add3(fixedpoint_t fxp1, fixedpoint_t fxp2, fixedpoint_t fxp3) {
// 	cnt_add++; 
// 	cnt_add++; 
	return fxp1+fxp2+fxp3;
}
/*
 fixedpoint_t fpm_Inc(fixedpoint_t fxp) {
	return fxp=fxp+FPM_FROMLONG(1);
}
*/
 fixedpoint_t fpm_Sub(fixedpoint_t fxp1, fixedpoint_t fxp2) {
// 	cnt_sub++;
	return fxp1-fxp2;
}
/*
 fixedpoint_t fpm_Dec(fixedpoint_t &fxp) {
	return fxp=fxp-FPM_FROMLONG(1);
}
*/
// KHUMEN TODO could be more effecient
fixedpoint_t fpm_Mul(fixedpoint_t fxp1, fixedpoint_t fxp2) {	
	bool needsign=false; 	
	word64 a,b; 
	word64 c,d;
	word64 rhh,rhl,rlh,rll;
	word64 res; 
//	cnt_mul++;
	if (fxp1<0) 
	 { needsign=true;
	    a=(word64)(-fxp1);	
	 	 } else a=(word64)(fxp1);
	if (fxp2<0) 
	 { if (needsign) needsign=false; else needsign=true;
	    b=(word64)(-fxp2);	
	 	 } else b=(word64)(fxp2); 	 
    c = (a>>32);
	d = (b>>32); 	 
	rhh = c*d;
	rhl = c*(b & 0xffffffff);
	rlh = (a & 0xffffffff)*d;
	rll = (a & 0xffffffff)*(b & 0xffffffff);
	res = (rhh<<32)+ rhl+rlh +(rll >>32);
	if ((rhh>>33)+((rhl+rlh)>>63)) res=0x7fffffffffffffff;
	if (needsign) 
	   return -(int64)(res);
	 else 
       return (int64)(res);
	
	
//return  fpm_FromFloat(fpm_ToDouble(fxp1)*fpm_ToDouble(fxp2));
	
}
# define MAD_F_FRACBITS		32
# define mad_f_intpart(x)	((x) >> MAD_F_FRACBITS)
# define MAD_F_MIN		((fixedpoint_t) -0x8000000000000000)
# define MAD_F_MAX		((fixedpoint_t) +0x7fffffffffffffff)

fixedpoint_t fpm_Div(fixedpoint_t fxp1, fixedpoint_t fxp2) {	
  fixedpoint_t q, r;
  unsigned int bits;
//  cnt_div++;
  if (fxp2==0) 
   { if (fxp1<0) 
   	    return MAD_F_MIN; 
   	 else 
   	    return MAD_F_MAX;
    }
  q = fpm_Abs(fxp1 / fxp2);
  if (fxp1 < 0) {
    fxp1 = -fxp1;
    fxp2 = -fxp2;
  }

  r = fxp1 % fxp2;

  if (fxp2 < 0) {
    fxp1 = -fxp1;
    fxp2 = -fxp2;
  }

  if (q > mad_f_intpart(MAD_F_MAX) &&
      !(q == -mad_f_intpart(MAD_F_MIN) && r == 0 && (fxp1 < 0) != (fxp2 < 0)))
    return 0;

  for (bits = MAD_F_FRACBITS; bits && r; --bits) {
    q <<= 1, r <<= 1;
    if (r >= fxp2)
      r -= fxp2, ++q;
  }

  if (2 * r >= fxp2)
    ++q;

  if ((fxp1 < 0) != (fxp2 < 0))
    q = -q;

  return q << bits;
 	
// 	return fpm_FromFloat(fpm_ToDouble(fxp1)/fpm_ToDouble(fxp2));
}

 fixedpoint_t fpm_DivInt(fixedpoint_t fxp1, long l) {
// 	cnt_div++;
	return fxp1/l;
}

 fixedpoint_t fpm_Abs(fixedpoint_t fxp) {
// 	cnt_abs++;
 	if (fxp>=0) {return fxp; } else {return -fxp; }
}

//TODO: could be more efficient
 fixedpoint_t fpm_Ceil(fixedpoint_t fxp) {
// 	cnt_ceil++;
	if (fxp&0xffffffff) {
		if (fxp<=0) return -(fixedpoint_t)((-fxp)&0xffffffff00000000);
		return (fxp&0xffffffff00000000)+FPM_FROMLONGC(1);
	}
	return fxp;
 	
// 	 return fpm_FromFloat(ceil(fpm_ToFloat(fxp)));
}
//TODO: could be more efficient
 fixedpoint_t fpm_Floor(fixedpoint_t fxp) {
// 	cnt_floor++;
	if (fxp&0xffffffff) {
		if (fxp<0) return -(fixedpoint_t)(((-fxp)&0xffffffff00000000)+FPM_FROMLONGC(1));
		return fxp&0xffffffff00000000;
	}
	return fxp; 	
// 	return fpm_FromFloat(floor(fpm_ToFloat(fxp)));
}

//TODO: Implement sqrt mathematically instead of converting to float and back
 fixedpoint_t fpm_Sqrt(fixedpoint_t fxp) {
// 	cnt_sqrt++;
	return fpm_FromFloat(sqrt(fpm_ToFloat(fxp)));
}

 fixedpoint_t fpm_Sqr(fixedpoint_t fxp) { 
	return fpm_Mul(fxp,fxp);
}

 fixedpoint_t fpm_Inv(fixedpoint_t fxp) {

	return fpm_Div(FPM_FROMLONGC(1),fxp);
}

//TODO: Calc trig functions (or lookup) instead of converting to float and back
//These take radians
 fixedpoint_t fpm_Sin(fixedpoint_t fxp) {
// 	cnt_sin++;
	return fpm_FromFloat(sin(fpm_ToFloat(fxp)));
}
 fixedpoint_t fpm_Cos(fixedpoint_t fxp) {
// 	cnt_cos++;
	return fpm_FromFloat(cos(fpm_ToFloat(fxp)));
}
 fixedpoint_t fpm_Tan(fixedpoint_t fxp) {
// 	cnt_tan++;
	return fpm_FromFloat(tan(fpm_ToFloat(fxp)));
}
 fixedpoint_t fpm_ATan(fixedpoint_t fxp) {
// 	cnt_atan++;
	return fpm_FromFloat(atan(fpm_ToFloat(fxp)));
}

//These take degrees
 fixedpoint_t fpm_SinDeg(fixedpoint_t fxp) {
	return fpm_Sin(fpm_DivInt(fpm_Mul(fxp, FPM_PI), 180));
}
 fixedpoint_t fpm_CosDeg(fixedpoint_t fxp) {
	return fpm_Cos(fpm_DivInt(fpm_Mul(fxp, FPM_PI), 180));
}
 fixedpoint_t fpm_TanDeg(fixedpoint_t fxp) {
	return fpm_Tan(fpm_DivInt(fpm_Mul(fxp, FPM_PI), 180));
}
 fixedpoint_t fpm_ATanDeg(fixedpoint_t fxp) {
	return fpm_ATan(fpm_DivInt(fpm_Mul(fxp, FPM_PI), 180));
}

/*********************************************/
/* 8.24 routines:                            */
/*********************************************/

#ifndef USEFPM
fixedpoint8_24_t fpm_FromFloat8_24(double f) {
#	if  defined(DEBUG)&&defined(FPM_VALIDATE)
	 if (f) ASSERT((fixedpoint_t) (f*16777216.0));
#	endif

	return (fixedpoint8_24_t) (f*16777216.0);
}

float fpm_ToFloat8_24(fixedpoint8_24_t fxp) {
	return (float)(fxp/16777216.0);
}

fixedpoint8_24_t fpm_FromLong8_24(long l) {
#	if  defined(DEBUG)&&defined(FPM_VALIDATE)
	 if (l>0) ASSERT((l<<24)>0);
	 else if (l<0) ASSERT((l<<24)<0);
#	endif

	return l<<24;
}

long fpm_ToLong8_24(fixedpoint8_24_t fxp) {
	if (fxp<0)
		return -((long)((fxp^0xffffffff)>>24)+1);
	else 
		return (fxp>>24)&0x000000ff;
}

fixedpoint8_24_t fpm_FromFixedPoint(fixedpoint_t fxp) {
	return fxp<<8;
}

fixedpoint_t fpm_ToFixedPoint(fixedpoint8_24_t fxp) {
	if (fxp<0)
		return -((long)((fxp^0xffffffff)>>8)+1);
	else 
		return (fxp>>8)&0x00ffffff;
}

fixedpoint8_24_t fpm_Add8_24(fixedpoint8_24_t fxp1, fixedpoint8_24_t fxp2) {
	return fxp1+fxp2;
}

fixedpoint8_24_t fpm_Add38_24(fixedpoint8_24_t fxp1, fixedpoint8_24_t fxp2, fixedpoint8_24_t fxp3) {
	return fxp1+fxp2+fxp3;
}
/*
 fixedpoint_t fpm_Inc(fixedpoint_t fxp) {
	return fxp=fxp+FPM_FROMLONG(1);
}
*/
fixedpoint8_24_t fpm_Sub8_24(fixedpoint8_24_t fxp1, fixedpoint8_24_t fxp2) {
	return fxp1-fxp2;
}
/*
 fixedpoint_t fpm_Dec(fixedpoint_t &fxp) {
	return fxp=fxp-FPM_FROMLONG(1);
}
*/
fixedpoint8_24_t fpm_Mul8_24(fixedpoint8_24_t fxp1, fixedpoint8_24_t fxp2) {
	 int64 tmp=fxp1;

#	if  defined(DEBUG)&&defined(FPM_VALIDATE)
	 fixedpoint8_24_t fxp;
	 tmp*=fxp2;
	 fxp=(fixedpoint8_24_t)(tmp>>24);
	 if (fxp1&&fxp2) {
		 ASSERT(fxp);
		 if ((fxp1>0&&fxp2>0)||(fxp1<0&&fxp2<0)) ASSERT(fxp>0);
		 else ASSERT(fxp<0);
	 }
	 return fxp;
#	endif

	tmp*=fxp2;
	return (fixedpoint_t)(tmp>>24);
}

fixedpoint_t fpm_MulMixed8_24(fixedpoint8_24_t fxp1, fixedpoint_t fxp2) {
	int64 tmp=fxp1;
	tmp*=fxp2;
	
	return (fixedpoint_t)(tmp>>24);
}

 fixedpoint8_24_t fpm_Div8_24(fixedpoint8_24_t fxp1, fixedpoint8_24_t fxp2) {
	int64 tmp=fxp1;
#	if  defined(DEBUG)&&defined(FPM_VALIDATE)
	//The purpose of this is to cause an exception in Windows CE
	//that will be handled right here by the debugger.  Otherwise a
	//divide by zero will be thrown which cannot be used to find
	//where the exception actually occured with the debugger.
	ASSERT(fxp2);
	if (!fxp2) 
		(*(int *)fxp2)=0;
#	endif
	tmp<<=24;
	return (fixedpoint8_24_t)(tmp/fxp2);
}

fixedpoint8_24_t fpm_DivInt8_24(fixedpoint8_24_t fxp1, long l) {
	return fxp1/l;
}

fixedpoint8_24_t fpm_DivInt64_8_24(fixedpoint8_24_t fxp1, int64 l) {
	return (fixedpoint8_24_t) (fxp1/l);
}

 fixedpoint8_24_t fpm_Abs8_24(fixedpoint8_24_t fxp) {
 	if (fxp>=0) {return fxp;} else { return -fxp;}
}

//TODO: could be more efficient
 fixedpoint8_24_t fpm_Ceil8_24(fixedpoint8_24_t fxp) {
	if (fxp&0x00ffffff) {
		if (fxp<=0) return (fxp&0xff000000);
		return (fxp&0xff000000)+FPM_FROMLONGC8_24(1);
	}
	return fxp;
}
//TODO: could be more efficient
 fixedpoint8_24_t fpm_Floor8_24(fixedpoint8_24_t fxp) {
	if (fxp&0x00ffffff) {
		if (fxp<0) return -(long)((fxp&0xff000000)+FPM_FROMLONG8_24(1));
		return fxp&0xff000000;
	}
	return fxp;
}

//TODO: Implement sqrt mathematically instead of converting to float and back
 fixedpoint8_24_t fpm_Sqrt8_24(fixedpoint8_24_t fxp) {
	return fpm_FromFloat8_24(sqrt(fpm_ToFloat8_24(fxp)));
}

 fixedpoint8_24_t fpm_Sqr8_24(fixedpoint8_24_t fxp) {
	return fpm_Mul8_24(fxp,fxp);
}

 fixedpoint8_24_t fpm_Inv8_24(fixedpoint8_24_t fxp) {
#	if  defined(DEBUG)&&defined(FPM_VALIDATE)
	 ASSERT(fxp);
#	endif

	return fpm_Div8_24(FPM_FROMLONGC8_24(1), fxp);
}
#endif