/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sv_user.c -- server code for moving users
/*********************************************************************
Copyright (c) 2001 Dan East
**********************************************************************

   Port of the original ID Software source code to Pocket PC,
   fixed point optimization, translation 
**********************************************************************
Copyright (c) 2006 Dr.Khumen Research Labs Ltd. 
          (c) 2006 Dmitry Gazko
**********************************************************************
   Port of PocketQuake source code to Symbian OS,
   fixed point translatation.       
**********************************************************************/


#include "quakedef.h"

#ifndef USEFPM
edict_t		*sv_player;
#else
edict_FPM_t	*sv_playerFPM;
#endif //USEFPM

extern	cvar_t	sv_friction;
cvar_t	sv_edgefriction = {"edgefriction", "2"};
extern	cvar_t	sv_stopspeed;

#ifndef  USEFPM
static	vec3_t		forward, right, up;
vec3_t	wishdir;
float	wishspeed;
float	*angles;
float	*origin;
float	*velocity;

// world
#else
float	*angles;
float	*origin;
float	*velocity;

static	vec3_FPM_t		forwardFPM, rightFPM, upFPM;
vec3_FPM_t	wishdirFPM;
fixedpoint_t	wishspeedFPM;

// world
//fixedpoint_t	*anglesFPM;
//fixedpoint_t	*originFPM;
//fixedpoint_t	*velocityFPM;
//vec3_FPM_t anglesFPM,originFPM,velocityFPM;
#endif

qboolean	onground;

#ifndef USEFPM
usercmd_t	cmd;
#else
usercmd_FPM_t cmdFPM;
#endif

cvar_t	sv_idealpitchscale = {"sv_idealpitchscale","0.8"};


/*
===============
SV_SetIdealPitch
===============
*/
#define	MAX_FORWARD	6
#ifndef USEFPM
void SV_SetIdealPitch (void)
{
	float	angleval, sinval, cosval;
	trace_t	tr;
	vec3_t	top, bottom;
	float	z[MAX_FORWARD];
	int		i, j;
	int		step, dir, steps;
//	char buf[256];

	if (!((int)sv_player->v.flags & FL_ONGROUND))
		return;
	
	angleval = (float)(sv_player->v.angles[YAW] * M_PI*2 / 360);
	sinval = (float)sin(angleval);
	cosval = (float)cos(angleval);

	//Dan:
// 	sprintf(buf, "%f %f ", sinval, cosval);
//	LogIt(buf);

	for (i=0 ; i<MAX_FORWARD ; i++)
	{
		top[0] = sv_player->v.origin[0] + cosval*(i+3)*12;
		top[1] = sv_player->v.origin[1] + sinval*(i+3)*12;
		top[2] = sv_player->v.origin[2] + sv_player->v.view_ofs[2];
		
		bottom[0] = top[0];
		bottom[1] = top[1];
		bottom[2] = top[2] - 160;
		
		tr = SV_Move (top, vec3_origin, vec3_origin, bottom, 1, sv_player);
		if (tr.allsolid)
			return;	// looking at a wall, leave ideal the way is was

		if (tr.fraction == 1)
			return;	// near a dropoff
		
		z[i] = top[2] + tr.fraction*(bottom[2]-top[2]);
	}

	dir = 0;
	steps = 0;
	for (j=1 ; j<i ; j++)
	{
		step = (int)(z[j] - z[j-1]);
		if (step > -ON_EPSILON && step < ON_EPSILON)
			continue;

		if (dir && ( step-dir > ON_EPSILON || step-dir < -ON_EPSILON ) )
			return;		// mixed changes

		steps++;	
		dir = step;
	}
	
	if (!dir)
	{
		sv_player->v.idealpitch = 0;
		return;
	}
	
	if (steps < 2)
		return;
	sv_player->v.idealpitch = -dir * sv_idealpitchscale.value;
}
#endif

#ifdef USEFPM
void SV_SetIdealPitchFPM (void)
{
	fixedpoint_t	angleval, sinval, cosval;
	trace_FPM_t		tr;
	vec3_FPM_t		top, bottom;
	fixedpoint_t	z[MAX_FORWARD];
	int				i, j;
	int				step, dir, steps;

	if (!((int)sv_playerFPM->v.flags & FL_ONGROUND))
		return;
	
	angleval = FPM_DIV(FPM_MUL(FPM_MUL(FPM_FROMFLOAT(sv_playerFPM->v.angles[YAW]), FPM_PI),FPM_FROMLONG(2)) , FPM_FROMLONG(360));
	sinval = FPM_SIN(angleval);
	cosval = FPM_COS(angleval);

	
	for (i=0 ; i<MAX_FORWARD ; i++)
	{
		top[0] = FPM_DIV(FPM_FROMFLOAT(sv_playerFPM->v.origin[0]), FPM_MUL(cosval,FPM_FROMLONG((i+3)*12)));
		top[1] = FPM_DIV(FPM_FROMFLOAT(sv_playerFPM->v.origin[1]), FPM_MUL(sinval,FPM_FROMLONG((i+3)*12)));
		top[2] = FPM_ADD(FPM_FROMFLOAT(sv_playerFPM->v.origin[2]), FPM_FROMFLOAT(sv_playerFPM->v.view_ofs[2]));
		
		bottom[0] = top[0];
		bottom[1] = top[1];
		bottom[2] = FPM_SUB(top[2],FPM_FROMLONG(160));
		
		tr = SV_MoveFPM (top, vec3_originFPM, vec3_originFPM, bottom, 1, sv_playerFPM);
		if (tr.allsolid)
			return;	// looking at a wall, leave ideal the way is was

		if (tr.fraction == FPM_FROMLONG(1))
			return;	// near a dropoff
		
			z[i] = FPM_ADD(top[2], FPM_MUL(tr.fraction,FPM_SUB(bottom[2],top[2])));
	}

	dir = 0;
	steps = 0;
	for (j=1 ; j<i ; j++)
	{
		step = FPM_TOLONG(FPM_SUB(z[j], z[j-1]));
		if (step > -ON_EPSILON && step < ON_EPSILON)
			continue;

		if (dir && ( step-dir > ON_EPSILON || step-dir < -ON_EPSILON ) )
			return;		// mixed changes

		steps++;	
		dir = step;
	}
	
	if (!dir)
	{
		sv_playerFPM->v.idealpitch = 0;
		return;
	}
	
	if (steps < 2)
		return;
	sv_playerFPM->v.idealpitch = -dir * sv_idealpitchscale.value;
}
#endif //USEFPM

/*
==================
SV_UserFriction

==================
*/
#ifdef USEFPM
void SV_UserFrictionFPM (void)
{
	float	*vel;
	fixedpoint_t	speed, newspeed, control;
	vec3_FPM_t	start, stop;
	fixedpoint_t	friction;
	trace_FPM_t	trace;
	
	vel = velocity;
	
	speed = FPM_SQRT(FPM_ADD(FPM_SQR(FPM_FROMFLOAT(vel[0])),FPM_SQR(FPM_FROMFLOAT(vel[1]))));
//	speed = FPM_SQRT(FPM_ADD(FPM_MUL(vel[0],vel[0]),FPM_MUL(vel[1],vel[1])));
	if (!speed)
		return;

// if the leading edge is over a dropoff, increase friction
	start[0] = stop[0] = FPM_ADD(FPM_FROMFLOAT(origin[0]) , FPM_MUL(FPM_DIV(FPM_FROMFLOAT(vel[0]),speed),FPM_FROMLONG(16)));
	start[1] = stop[1] = FPM_ADD(FPM_FROMFLOAT(origin[1]) , FPM_MUL(FPM_DIV(FPM_FROMFLOAT(vel[1]),speed),FPM_FROMLONG(16)));
	start[2] = FPM_ADD(FPM_FROMFLOAT(origin[2]) , FPM_FROMFLOAT(sv_playerFPM->v.mins[2]));
	stop[2] = FPM_SUB(start[2] , FPM_FROMLONG(34));

	trace = SV_MoveFPM (start, vec3_originFPM, vec3_originFPM, stop, true, sv_playerFPM);

	if (trace.fraction == FPM_FROMLONG(1))
		friction = FPM_FROMFLOAT(sv_friction.value*sv_edgefriction.value);
	else
		friction = FPM_FROMFLOAT(sv_friction.value);

// apply friction	
	control = speed < FPM_FROMFLOAT(sv_stopspeed.value) ? FPM_FROMFLOAT(sv_stopspeed.value) : speed;
	newspeed = FPM_SUB(speed , FPM_MUL(FPM_MUL(FPM_FROMFLOAT(host_frametime),control),friction));
	
	if (newspeed < 0)
		newspeed = 0;
	newspeed = FPM_DIV(newspeed,speed);

	vel[0] = vel[0] * FPM_TOFLOAT(newspeed);
	vel[1] = vel[1] * FPM_TOFLOAT(newspeed);
	vel[2] = vel[2] * FPM_TOFLOAT(newspeed);
	
}
#endif
/*
==============
SV_Accelerate
==============
*/
cvar_t	sv_maxspeed = {"sv_maxspeed", "320", false, true};
cvar_t	sv_accelerate = {"sv_accelerate", "10"};
#if 0
void SV_Accelerate (vec3_t wishvel)
{
	int			i;
	float		addspeed, accelspeed;
	vec3_t		pushvec;

	if (wishspeed == 0)
		return;

	VectorSubtract (wishvel, velocity, pushvec);
	addspeed = VectorNormalize (pushvec);

	accelspeed = sv_accelerate.value*host_frametime*addspeed;
	if (accelspeed > addspeed)
		accelspeed = addspeed;
	
	for (i=0 ; i<3 ; i++)
		velocity[i] += accelspeed*pushvec[i];	
}
#endif
#ifndef USEFPM
void SV_Accelerate (void)
{
	int			i;
	float		addspeed, accelspeed, currentspeed;

	currentspeed = DotProduct (velocity, wishdir);
	addspeed = wishspeed - currentspeed;
	if (addspeed <= 0)
		return;
	accelspeed = (float)(sv_accelerate.value*host_frametime*wishspeed);
	if (accelspeed > addspeed)
		accelspeed = addspeed;
	
	for (i=0 ; i<3 ; i++)
		velocity[i] += accelspeed*wishdir[i];	
}
#else
void SV_AccelerateFPM (void)
{
	int			i;
	fixedpoint_t addspeed, accelspeed, currentspeed;
	vec3_FPM_t velocityFPM;
    VectorToFPM(velocity,velocityFPM);
	currentspeed = DotProductFPM (velocityFPM, wishdirFPM);
	addspeed = FPM_SUB(wishspeedFPM , currentspeed);
	if (addspeed <= 0)
		return;
	accelspeed = FPM_FROMFLOAT(sv_accelerate.value*host_frametime*FPM_TOFLOAT(wishspeedFPM));
	if (accelspeed > addspeed)
		accelspeed = addspeed;
	
	for (i=0 ; i<3 ; i++)
		velocity[i] += FPM_TOFLOAT(FPM_MUL(accelspeed,wishdirFPM[i]));	
} 
#endif

#ifndef USEFPM
void SV_AirAccelerate (vec3_t wishveloc)
{
	int			i;
	float		addspeed, wishspd, accelspeed, currentspeed;
		
	wishspd = VectorNormalize (wishveloc);
	if (wishspd > 30)
		wishspd = 30;
	currentspeed = DotProduct (velocity, wishveloc);
	addspeed = wishspd - currentspeed;
	if (addspeed <= 0)
		return;
//	accelspeed = sv_accelerate.value * host_frametime;
	accelspeed = (float)(sv_accelerate.value*wishspeed * host_frametime);
	if (accelspeed > addspeed)
		accelspeed = addspeed;
	
	for (i=0 ; i<3 ; i++)
		velocity[i] += accelspeed*wishveloc[i];	
}
#else
void SV_AirAccelerateFPM (vec3_FPM_t wishveloc)
{
	int			i;
	fixedpoint_t	addspeed, wishspd, accelspeed, currentspeed;
	vec3_FPM_t velocityFPM;
		
	wishspd = VectorNormalizeFPM (wishveloc);
	if (wishspd > FPM_FROMLONG(30))
		wishspd = FPM_FROMLONG(30);
	VectorToFPM(velocity,velocityFPM);	
	currentspeed = DotProductFPM (velocityFPM, wishveloc);
	addspeed = FPM_SUB(wishspd , currentspeed);
	if (addspeed <= 0)
		return;
//	accelspeed = sv_accelerate.value * host_frametime;
	accelspeed = FPM_FROMFLOAT(sv_accelerate.value*FPM_TOFLOAT(wishspeedFPM) * host_frametime);
	if (accelspeed > addspeed)
		accelspeed = addspeed;
	
	for (i=0 ; i<3 ; i++)
		velocity[i] += FPM_TOFLOAT(FPM_MUL(accelspeed,wishveloc[i]));	
}

#endif



void DropPunchAngle (void)
{
	float	len;
	
	len = VectorNormalize (sv_playerFPM->v.punchangle);
	
	len -= (float)(10*host_frametime);
	if (len < 0)
		len = 0;
	VectorScale (sv_playerFPM->v.punchangle, len, sv_playerFPM->v.punchangle);
}

/*
===================
SV_WaterMove

===================
*/
#ifndef USEFPM
void SV_WaterMove (void)
{
	int		i;
	vec3_t	wishvel;
	float	speed, newspeed, wishspeed, addspeed, accelspeed;

//
// user intentions
//
	AngleVectors (sv_player->v.v_angle, forward, right, up);

	for (i=0 ; i<3 ; i++)
		wishvel[i] = forward[i]*cmd.forwardmove + right[i]*cmd.sidemove;

	if (!cmd.forwardmove && !cmd.sidemove && !cmd.upmove)
		wishvel[2] -= 60;		// drift towards bottom
	else
		wishvel[2] += cmd.upmove;

	wishspeed = Length(wishvel);
	if (wishspeed > sv_maxspeed.value)
	{
		VectorScale (wishvel, sv_maxspeed.value/wishspeed, wishvel);
		wishspeed = sv_maxspeed.value;
	}
	wishspeed *= (float)0.7;

//
// water friction
//
	speed = Length (velocity);
	if (speed)
	{
		newspeed = (float)(speed - host_frametime * speed * sv_friction.value);
		if (newspeed < 0)
			newspeed = 0;	
		VectorScale (velocity, newspeed/speed, velocity);
	}
	else
		newspeed = 0;
	
//
// water acceleration
//
	if (!wishspeed)
		return;

	addspeed = wishspeed - newspeed;
	if (addspeed <= 0)
		return;

	VectorNormalize (wishvel);
	accelspeed = (float)(sv_accelerate.value * wishspeed * host_frametime);
	if (accelspeed > addspeed)
		accelspeed = addspeed;

	for (i=0 ; i<3 ; i++)
		velocity[i] += accelspeed * wishvel[i];
}
#else
void SV_WaterMoveFPM (void)
{
	int		i;
	vec3_FPM_t	wishvel,angl;
	fixedpoint_t	speed, newspeed, wishspeed, addspeed, accelspeed;
	vec3_FPM_t velocityFPM;

//
// user intentions
//  
        VectorToFPM(sv_playerFPM->v.v_angle,angl)
	AngleVectorsFPM (angl, forwardFPM, rightFPM, upFPM);

	for (i=0 ; i<3 ; i++)
		wishvel[i] = FPM_ADD(FPM_MUL(forwardFPM[i],cmdFPM.forwardmove) , FPM_MUL(rightFPM[i],cmdFPM.sidemove));

	if (!cmdFPM.forwardmove && !cmdFPM.sidemove && !cmdFPM.upmove)
		wishvel[2] = FPM_SUB(wishvel[2],FPM_FROMLONG(60));		// drift towards bottom
	else
		wishvel[2] = FPM_ADD(wishvel[2],cmdFPM.upmove);

	wishspeed = LengthFPM(wishvel);
	if (wishspeed > FPM_FROMFLOAT(sv_maxspeed.value))
	{
		VectorScaleFPM (wishvel, FPM_DIV(FPM_FROMFLOAT(sv_maxspeed.value),wishspeed), wishvel);
		wishspeed = FPM_FROMFLOAT(sv_maxspeed.value);
	}
	wishspeed = FPM_MUL(wishspeed,FPM_FROMFLOAT(0.7));

//
// water friction
//  
  VectorToFPM(velocity,velocityFPM);
	speed = LengthFPM (velocityFPM);
	if (speed)
	{
		newspeed = FPM_SUB(speed , FPM_FROMFLOAT(host_frametime * FPM_TOFLOAT(speed) * sv_friction.value));
		if (newspeed < 0)
			newspeed = 0;	
		VectorScale(velocity,FPM_TOFLOAT(FPM_DIV(newspeed,speed)),velocity);	
	}
	else
		newspeed = 0;
	
//
// water acceleration
//
	if (!wishspeed)
		return;

	addspeed = FPM_SUB(wishspeed,newspeed);
	if (addspeed <= 0)
		return;

	VectorNormalizeFPM (wishvel);
	accelspeed = FPM_FROMFLOAT(sv_accelerate.value * FPM_TOFLOAT(wishspeed) * host_frametime);
	if (accelspeed > addspeed)
		accelspeed = addspeed;

	for (i=0 ; i<3 ; i++)
		velocity[i] += FPM_TOFLOAT(FPM_MUL(accelspeed , wishvel[i]));
}

#endif
#ifndef USEFPM
void SV_WaterJump (void)
{
	if (sv.time > sv_player->v.teleport_time
	|| !sv_player->v.waterlevel)
	{
		sv_player->v.flags = (float)((int)sv_player->v.flags & ~FL_WATERJUMP);
		sv_player->v.teleport_time = 0;
	}
	sv_player->v.velocity[0] = sv_player->v.movedir[0];
	sv_player->v.velocity[1] = sv_player->v.movedir[1];
}
#else
void SV_WaterJumpFPM (void)
{
	if (svFPM.time > sv_playerFPM->v.teleport_time
	|| !sv_playerFPM->v.waterlevel)
	{
		sv_playerFPM->v.flags = (float)((int)sv_playerFPM->v.flags & ~FL_WATERJUMP);
		sv_playerFPM->v.teleport_time = 0;
	}
	sv_playerFPM->v.velocity[0] = sv_playerFPM->v.movedir[0];
	sv_playerFPM->v.velocity[1] = sv_playerFPM->v.movedir[1];
}
#endif

/*
===================
SV_AirMove

===================
*/
#ifndef USEFPM
void SV_AirMove (void)
{
	int			i;
	vec3_t		wishvel;
	float		fmove, smove;

	AngleVectors (sv_player->v.angles, forward, right, up);

	fmove = cmd.forwardmove;
	smove = cmd.sidemove;
	
// hack to not let you back into teleporter
	if (sv.time < sv_player->v.teleport_time && fmove < 0)
		fmove = 0;
		
	for (i=0 ; i<3 ; i++)
		wishvel[i] = forward[i]*fmove + right[i]*smove;

	if ( (int)sv_player->v.movetype != MOVETYPE_WALK)
		wishvel[2] = cmd.upmove;
	else
		wishvel[2] = 0;

	VectorCopy (wishvel, wishdir);
	wishspeed = VectorNormalize(wishdir);
	if (wishspeed > sv_maxspeed.value)
	{
		VectorScale (wishvel, sv_maxspeed.value/wishspeed, wishvel);
		wishspeed = sv_maxspeed.value;
	}
	
	if ( sv_player->v.movetype == MOVETYPE_NOCLIP)
	{	// noclip
		VectorCopy (wishvel, velocity);
	}
	else if ( onground )
	{
		SV_UserFriction ();
		SV_Accelerate ();
	}
	else
	{	// not on ground, so little effect on velocity
		SV_AirAccelerate (wishvel);
	}		
}
#else
void SV_AirMoveFPM (void)
{
	int			i;
	vec3_FPM_t		wishvel,angl;
	fixedpoint_t		fmove, smove;

    VectorToFPM(sv_playerFPM->v.angles,angl);
	AngleVectorsFPM (angl, forwardFPM, rightFPM, upFPM);

	fmove = cmdFPM.forwardmove;
	smove = cmdFPM.sidemove;
	
// hack to not let you back into teleporter
	if (svFPM.time < sv_playerFPM->v.teleport_time && fmove < 0)
		fmove = 0;
		
	for (i=0 ; i<3 ; i++)
		wishvel[i] = FPM_ADD(FPM_MUL(forwardFPM[i],fmove),FPM_MUL(rightFPM[i],smove));

	if ( (int)sv_playerFPM->v.movetype != MOVETYPE_WALK)
		wishvel[2] = cmdFPM.upmove;
	else
		wishvel[2] = 0;

	VectorCopy (wishvel, wishdirFPM);
	wishspeedFPM = VectorNormalizeFPM(wishdirFPM);
	if (wishspeedFPM > FPM_FROMFLOAT(sv_maxspeed.value))
	{
		VectorScaleFPM (wishvel, FPM_DIV(FPM_FROMFLOAT(sv_maxspeed.value),wishspeedFPM), wishvel);
		wishspeedFPM = FPM_FROMFLOAT(sv_maxspeed.value);
	}
	
	if ( sv_playerFPM->v.movetype == MOVETYPE_NOCLIP)
	{	// noclip
		FPMToVector(wishvel, velocity);
	}
	else if ( onground )
	{
		SV_UserFrictionFPM ();
		SV_AccelerateFPM ();
	}
	else
	{	// not on ground, so little effect on velocity
		SV_AirAccelerateFPM (wishvel);
	}		
}
#endif
/*
===================
SV_ClientThink

the move fields specify an intended velocity in pix/sec
the angle fields specify an exact angular motion in degrees
===================
*/
#ifndef USEFPM
void SV_ClientThink (void)
{
	vec3_t		v_angle;

	if (sv_player->v.movetype == MOVETYPE_NONE)
		return;
	
	onground = (int)sv_player->v.flags & FL_ONGROUND;

	origin = sv_player->v.origin;
	velocity = sv_player->v.velocity;

	DropPunchAngle ();
	
//
// if dead, behave differently
//
	if (sv_player->v.health <= 0)
		return;

//
// angles
// show 1/3 the pitch angle and all the roll angle
	cmd = host_client->cmd;
	angles = sv_player->v.angles;
	
	VectorAdd (sv_player->v.v_angle, sv_player->v.punchangle, v_angle);
	angles[ROLL] = V_CalcRoll (sv_player->v.angles, sv_player->v.velocity)*4;
	if (!sv_player->v.fixangle)
	{
		angles[PITCH] = -v_angle[PITCH]/3;
		angles[YAW] = v_angle[YAW];
	}

	if ( (int)sv_player->v.flags & FL_WATERJUMP )
	{
		SV_WaterJump ();
		return;
	}
//
// walk
//
	if ( (sv_player->v.waterlevel >= 2)
	&& (sv_player->v.movetype != MOVETYPE_NOCLIP) )
	{
		SV_WaterMove ();
		return;
	}

	SV_AirMove ();	
}

#else
void SV_ClientThinkFPM (void)
{
	vec3_FPM_t		v_angle;
	vec3_FPM_t      an,vel; 
	vec3_t  v_angle1;

	if (sv_playerFPM->v.movetype == MOVETYPE_NONE)
		return;
	
	onground = (int)sv_playerFPM->v.flags & FL_ONGROUND;

     origin=sv_playerFPM->v.origin;
     velocity=sv_playerFPM->v.velocity;
//    VectorToFPM(sv_playerFPM->v.origin,originFPM);
//    VectorToFPM(sv_playerFPM->v.velocity,velocityFPM);

	DropPunchAngle ();
	
//
// if dead, behave differently
//
	if (sv_playerFPM->v.health <= 0)
		return;

//
// angles
// show 1/3 the pitch angle and all the roll angle
	cmdFPM = host_clientFPM->cmd;
	angles=sv_playerFPM->v.angles;
//	VectorToFPM(sv_playerFPM->v.angles,anglesFPM);
	
	VectorAdd (sv_playerFPM->v.v_angle, sv_playerFPM->v.punchangle, v_angle1);
	VectorToFPM(v_angle1,v_angle);
	VectorToFPM(sv_playerFPM->v.angles,an);
	VectorToFPM(sv_playerFPM->v.velocity,vel);
	angles[ROLL] = FPM_TOFLOAT(FPM_MUL(V_CalcRollFPM (an, vel),FPM_FROMLONG(4)));
	if (!sv_playerFPM->v.fixangle)
	{
		angles[PITCH] = FPM_TOFLOAT(FPM_DIV(v_angle[PITCH],FPM_FROMLONG(-3)));
		angles[YAW] = FPM_TOFLOAT(v_angle[YAW]);
	}
	if ( (int)sv_playerFPM->v.flags & FL_WATERJUMP )
	{
		SV_WaterJumpFPM ();
		return;
	}
//
// walk
//
	if ( (sv_playerFPM->v.waterlevel >= 2)
	&& (sv_playerFPM->v.movetype != MOVETYPE_NOCLIP) )
	{
		SV_WaterMoveFPM ();
		return;
	}

	SV_AirMoveFPM ();	
}

#endif

/*
===================
SV_ReadClientMove
===================
*/
#ifndef USEFPM
void SV_ReadClientMove (usercmd_t *move)
{
	int		i;
	vec3_t	angle;
	int		bits;
	
// read ping time
	host_client->ping_times[host_client->num_pings%NUM_PING_TIMES]
		= (float)(sv.time - MSG_ReadFloat ());
	host_client->num_pings++;

// read current angles	
	for (i=0 ; i<3 ; i++)
		angle[i] = MSG_ReadAngle ();

	VectorCopy (angle, host_client->edict->v.v_angle);
		
// read movement
	move->forwardmove = (float)MSG_ReadShort ();
	move->sidemove = (float)MSG_ReadShort ();
	move->upmove = (float)MSG_ReadShort ();
	
// read buttons
	bits = MSG_ReadByte ();
	host_client->edict->v.button0 = (float)(bits & 1);
	host_client->edict->v.button2 = (float)((bits & 2)>>1);

	i = MSG_ReadByte ();
	if (i)
		host_client->edict->v.impulse = (float)i;

#ifdef QUAKE2
// read light level
	host_client->edict->v.light_level = MSG_ReadByte ();
#endif
}
#else
void SV_ReadClientMoveFPM (usercmd_FPM_t *move)
{
	int		i;
	vec3_FPM_t	angle;
	vec3_t an;
	int		bits;
	
// read ping time
	host_clientFPM->ping_times[host_clientFPM->num_pings%NUM_PING_TIMES]
		= (float)(svFPM.time - MSG_ReadFloat ());
	host_clientFPM->num_pings++;

// read current angles	
	for (i=0 ; i<3 ; i++)
		angle[i] = MSG_ReadAngleFPM ();
  FPMToVector(angle,an);
	VectorCopy (an, host_clientFPM->edict->v.v_angle);
		
// read movement
	move->forwardmove = FPM_FROMLONG(MSG_ReadShort ());
	move->sidemove = FPM_FROMLONG(MSG_ReadShort ());
	move->upmove = FPM_FROMLONG(MSG_ReadShort ());
	
// read buttons
	bits = MSG_ReadByte ();
	host_clientFPM->edict->v.button0 = (float)(bits & 1);
	host_clientFPM->edict->v.button2 = (float)((bits & 2)>>1);

	i = MSG_ReadByte ();
	if (i)
		host_clientFPM->edict->v.impulse = (float)i;

#ifdef QUAKE2
// read light level
	host_clientFPM->edict->v.light_level = MSG_ReadByte ();
#endif
}

#endif

/*
===================
SV_ReadClientMessage

Returns false if the client should be killed
===================
*/
#ifndef USEFPM
qboolean SV_ReadClientMessage (void)
{
	int		ret;
	int		cmd;
	char		*s;
	
	do
	{
nextmsg:
		ret = NET_GetMessage (host_client->netconnection);
		if (ret == -1)
		{
			Sys_Printf ("SV_ReadClientMessage: NET_GetMessage failed\n");
			return false;
		}
		if (!ret)
			return true;
					
		MSG_BeginReading ();
		
		while (1)
		{
			if (!host_client->active)
				return false;	// a command caused an error

			if (msg_badread)
			{
				Sys_Printf ("SV_ReadClientMessage: badread\n");
				return false;
			}	
	
			cmd = MSG_ReadChar ();
			
			switch (cmd)
			{
			case -1:
				goto nextmsg;		// end of message
				
			default:
				Sys_Printf ("SV_ReadClientMessage: unknown command char\n");
				return false;
							
			case clc_nop:
//				Sys_Printf ("clc_nop\n");
				break;
				
			case clc_stringcmd:	
				s = MSG_ReadString ();
				if (host_client->privileged)
					ret = 2;
				else
					ret = 0;
				if (Q_strncasecmp(s, "status", 6) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "god", 3) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "notarget", 8) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "fly", 3) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "name", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "noclip", 6) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "say", 3) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "say_team", 8) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "tell", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "color", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "kill", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "pause", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "spawn", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "begin", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "prespawn", 8) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "kick", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "ping", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "give", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "ban", 3) == 0)
					ret = 1;
				if (ret == 2)
					Cbuf_InsertText (s);
				else if (ret == 1)
					Cmd_ExecuteString (s, src_client);
				else
					Con_DPrintf("%s tried to %s\n", host_client->name, s);
				break;
				
			case clc_disconnect:
//				Sys_Printf ("SV_ReadClientMessage: client disconnected\n");
				return false;
			
			case clc_move:
				SV_ReadClientMove (&host_client->cmd);
				break;
			}
		}
	} while (ret == 1);
	
	return true;
}
#else
qboolean SV_ReadClientMessageFPM (void)
{
	int		ret;
	int		cmd;
	char		*s;
	
	do
	{
nextmsg:
		ret = NET_GetMessage (host_clientFPM->netconnection);
		if (ret == -1)
		{
			Sys_Printf ("SV_ReadClientMessage: NET_GetMessage failed\n");
			return false;
		}
		if (!ret)
			return true;
					
		MSG_BeginReading ();
		
		while (1)
		{
			if (!host_clientFPM->active)
				return false;	// a command caused an error

			if (msg_badread)
			{
				Sys_Printf ("SV_ReadClientMessage: badread\n");
				return false;
			}	
	
			cmd = MSG_ReadChar ();
			
			switch (cmd)
			{
			case -1:
				goto nextmsg;		// end of message
				
			default:
				Sys_Printf ("SV_ReadClientMessage: unknown command char\n");
				return false;
							
			case clc_nop:
//				Sys_Printf ("clc_nop\n");
				break;
				
			case clc_stringcmd:	
				s = MSG_ReadString ();
				if (host_clientFPM->privileged)
					ret = 2;
				else
					ret = 0;
				if (Q_strncasecmp(s, "status", 6) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "god", 3) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "notarget", 8) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "fly", 3) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "name", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "noclip", 6) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "say", 3) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "say_team", 8) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "tell", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "color", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "kill", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "pause", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "spawn", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "begin", 5) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "prespawn", 8) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "kick", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "ping", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "give", 4) == 0)
					ret = 1;
				else if (Q_strncasecmp(s, "ban", 3) == 0)
					ret = 1;
				if (ret == 2)
					Cbuf_InsertText (s);
				else if (ret == 1)
					Cmd_ExecuteString (s, src_client);
				else
					Con_DPrintf("%s tried to %s\n", host_clientFPM->name, s);
				break;
				
			case clc_disconnect:
//				Sys_Printf ("SV_ReadClientMessage: client disconnected\n");
				return false;
			
			case clc_move:
				SV_ReadClientMoveFPM (&host_clientFPM->cmd);
				break;
			}
		}
	} while (ret == 1);
	
	return true;
}

#endif

/*
==================
SV_RunClients
==================
*/
#ifndef USEFPM
void SV_RunClients (void)
{
	int				i;
	
	for (i=0, host_client = svs.clients ; i<svs.maxclients ; i++, host_client++)
	{
		if (!host_client->active)
			continue;
	
		sv_player = host_client->edict;

		if (!SV_ReadClientMessage ())
		{
			SV_DropClient (false);	// client misbehaved...
			continue;
		}

		if (!host_client->spawned)
		{
		// clear client movement until a new packet is received
			memset (&host_client->cmd, 0, sizeof(host_client->cmd));
			continue;
		}

// always pause in single player if in console or menus
		if (!sv.paused && (svs.maxclients > 1 || key_dest == key_game) )
			SV_ClientThink ();
	}
}
#else
void SV_RunClientsFPM (void)
{
	int				i;
	
	for (i=0, host_clientFPM = svsFPM.clients ; i<svsFPM.maxclients ; i++, host_clientFPM++)
	{
		if (!host_clientFPM->active)
			continue;
	
		sv_playerFPM = host_clientFPM->edict;

		if (!SV_ReadClientMessageFPM ())
		{
			SV_DropClientFPM (false);	// client misbehaved...
			continue;
		}

		if (!host_clientFPM->spawned)
		{
		// clear client movement until a new packet is received
			memset (&host_clientFPM->cmd, 0, sizeof(host_clientFPM->cmd));
			continue;
		}

// always pause in single player if in console or menus
		if (!svFPM.paused && (svsFPM.maxclients > 1 || key_dest == key_game) )
			SV_ClientThinkFPM();
	}
}
#endif
